# -*- coding: utf-8 -*-

from libcomxml.core import XmlModel, XmlField
from .utils import FacturaeUtils
from signxml import XMLSigner, XMLVerifier
from xml.etree import ElementTree


class Extensions(XmlModel):

    _sort_order = ('extensions', 'utilities')

    def __init__(self):
        self.extensions = XmlField('Extensions')
        self.utilities = []
        super(Extensions, self).__init__('Extensions', 'extensions')

# 2.2


class UtilitiesExtensionsELE(XmlModel):

    _sort_order = ('utilitiesextensionsele', 'version', 'utilitiesfacturaresumen', 'utilitiesextensionele')

    def __init__(self):
        nsmap = {'utilities': 'http://www.facturae.es/Facturae/Extensions/Utilities'}

        self.utilitiesextensionsele = XmlField('UtilitiesExtensionsELE', namespace=nsmap['utilities'])
        self.version = XmlField('Version')
        self.utilitiesfacturaresumen = UtilitiesFacturaResumen()
        self.utilitiesextensionele = UtilitiesExtensionELE()
        super(UtilitiesExtensionsELE, self).__init__('UtilitiesExtensionsELE', 'utilitiesextensionsele')

# 2.2


class UtilitiesFacturaResumen(XmlModel):

    _sort_order = ('utilitiesfacturaresumen','facturaresumen', 'conceptosfr', 'totalantesimpuestos', 'impuestosfr')

    def __init__(self):
        self.utilitiesfacturaresumen = XmlField('UtilitiesFacturaResumen')
        self.facturaresumen = XmlField('FacturaResumen')
        self.conceptosfr = ConceptosFR()
        self.totalantesimpuestos = XmlField('TotalAntesImpuestos')
        self.impuestosfr = ImpuestosFR()
        super(UtilitiesFacturaResumen, self).__init__('UtilitiesFacturaResumen', 'utilitiesfacturaresumen')

# 2.2


class ConceptosFR(XmlModel):

    _sort_order = ('conceptosfr', 'conceptofr')

    def __init__(self):
        self.conceptosfr = XmlField('ConceptosFR')
        self.conceptofr = []
        super(ConceptosFR, self).__init__('ConceptosFR', 'conceptosfr')

# 2.2


class ConceptoFR(XmlModel):

    _sort_order = ('conceptofr', 'conceptofacturadofr', 'tipoconceptofacturadofr', 'importe')

    def __init__(self):
        self.conceptofr = XmlField('ConceptoFR')
        self.conceptofacturadofr = XmlField('ConceptoFacturadoFR')
        self.tipoconceptofacturadofr = XmlField('TipoConceptoFacturadoFR')
        self.importe = XmlField('Importe')
        super(ConceptoFR, self).__init__('ConceptoFR', 'conceptofr')

# 2.2


class ImpuestosFR(XmlModel):

    _sort_order = ('impuestosfr', 'impuesto')

    def __init__(self):
        self.impuestosfr = XmlField('ImpuestosFR')
        self.impuesto = []
        super(ImpuestosFR, self).__init__('ImpuestosFR', 'impuestosfr')

# 2.2


class Impuesto(XmlModel):

    _sort_order = ('impuesto', 'base', 'tipoimpuesto', 'tipoimpositivo', 'tipooperacion', 'importe')

    def __init__(self):
        self.impuesto = XmlField('Impuesto')
        self.base = XmlField('Base')
        self.tipoimpuesto = XmlField('TipoImpuesto')
        self.tipoimpositivo = XmlField('TipoImpositivo')
        self.tipooperacion = XmlField('TipoOperacion')
        self.importe = XmlField('Importe')
        super(Impuesto, self).__init__('Impuesto', 'impuesto')

# 2.2


class UtilitiesExtensionELE(XmlModel):

    _sort_order = ('utilitiesextensionele', 'tipoextension', 'datosdelsuministro', 'utilitiesmedida', 'utilitiesdesgloseconceptosfactura'
                   , 'utilitieshistoricoconsumos', 'utilitiesmensajes', 'utilitiesatrasociado', 'utilitiescurva')

    def __init__(self):
        self.utilitiesextensionele = XmlField('UtilitiesExtensionELE')
        self.tipoextension = XmlField('TipoExtension')
        self.datosdelsuministro = DatosDelSuministro()
        self.utilitiesmedida = UtilitiesMedida()
        self.utilitiesdesgloseconceptosfactura = UtilitiesDesgloseConceptosFactura()
        self.utilitieshistoricoconsumos = UtilitiesHistoricoConsumos()
        self.utilitiesmensajes = UtilitiesMensajes()
        self.utilitiesatrasociado = UtilitiesAtrAsociado()
        self.utilitiescurva = UtilitiesCurva()
        super(UtilitiesExtensionELE, self).__init__('UtilitiesExtensionELE', 'utilitiesextensionele')

# 2.2


class DatosDelSuministro(XmlModel):

    _sort_order = ('datosdelsuministro', 'cups', 'direccionsuministro', 'contrato', 'potenciascontratadas',
                   'datosdistribuidora', 'referenciaslegal', 'tarifa', 'agrupacion', 'referencia')

    def __init__(self):
        self.datosdelsuministro = XmlField('DatosDelSuministro')
        self.cups = XmlField('CUPS')
        self.direccionsuministro = DireccionSuministro()
        self.contrato = Contrato()
        self.potenciascontratadas = PotenciasContratadas()
        self.datosdistribuidora = DatosDistribuidora()
        self.referenciaslegal = ReferenciasLegal()
        self.tarifa = Tarifa()
        self.agrupacion = Agrupacion()
        self.referencia = XmlField('Referencia')
        super(DatosDelSuministro, self).__init__('DatosDelSuministro', 'datosdelsuministro')

# 2.2


class DireccionSuministro(XmlModel):

    _sort_order = ('direccionsuministro', 'direccion', 'codigopostal', 'poblacion', 'provincia', 'pais')

    def __init__(self):
        self.direccionsuministro = XmlField('DireccionSuministro')
        self.direccion = XmlField('Direccion')
        self.codigopostal = XmlField('CodigoPostal')
        self.poblacion = XmlField('Poblacion')
        self.provincia = XmlField('Provincia')
        self.pais = XmlField('Pais')
        super(DireccionSuministro, self).__init__('DireccionSuministro', 'direccionsuministro')

# 2.2


class Contrato(XmlModel):

    _sort_order = ('contrato', 'refcontratoempresa', 'fechafincontrato')

    def __init__(self):
        self.contrato = XmlField('Contrato')
        self.refcontratoempresa = XmlField('RefContratoEmpresa')
        self.fechafincontrato = XmlField('FechaFinContrato')
        super(Contrato, self).__init__('Contrato', 'contrato')

# 2.2


class PotenciasContratadas(XmlModel):

    _sort_order = ('potenciascontratadas', 'potenciacontratada')

    def __init__(self):
        self.potenciascontratadas = XmlField('PotenciasContratadas')
        self.potenciacontratada = []
        super(PotenciasContratadas, self).__init__('PotenciasContratadas', 'potenciascontratadas')

# 2.2


class PotenciaContratada(XmlModel):

    _sort_order = ('potenciacontratada', 'periodo', 'valor', 'unidadmedida', 'fechadesde', 'fechahasta')

    def __init__(self):
        self.potenciacontratada = XmlField('PotenciaContratada')
        self.periodo = XmlField('Periodo')
        self.valor = XmlField('Valor')
        self.unidadmedida = XmlField('UnidadMedida')
        self.fechadesde = XmlField('FechaDesde')
        self.fechahasta = XmlField('FechaHasta')
        super(PotenciaContratada, self).__init__('PotenciaContratada', 'potenciacontratada')

# 2.2


class DatosDistribuidora(XmlModel):

    _sort_order = ('datosdistribuidora', 'distribuidora', 'telefonoaverias')

    def __init__(self):
        self.datosdistribuidora = XmlField('DatosDistribuidora')
        self.distribuidora = XmlField('Distribuidora')
        self.telefonoaverias = XmlField('TelefonoAverias')
        super(DatosDistribuidora, self).__init__('DatosDistribuidora', 'datosdistribuidora')

# 2.2


class ReferenciasLegal(XmlModel):

    _sort_order = ('referenciaslegal', 'referencialegal')

    def __init__(self):
        self.referenciaslegal = XmlField('DatosDistribuidora')
        self.referencialegal = ReferenciaLegal()
        super(ReferenciasLegal, self).__init__('ReferenciasLegal', 'referenciaslegal')

# 2.2


class ReferenciaLegal(XmlModel):

    _sort_order = ('referencialegal', 'boeboca', 'fechaboeboca')

    def __init__(self):
        self.referencialegal = XmlField('ReferenciaLegal')
        self.boeboca = XmlField('BOEBOCA')
        self.fechaboeboca = XmlField('FechaBOEBOCA')
        super(ReferenciaLegal, self).__init__('ReferenciaLegal', 'referencialegal')

# 2.2


class Tarifa(XmlModel):

    _sort_order = ('tarifa', 'CodigoTarifaProducto', 'AltaMedidaEnBaja')

    def __init__(self):
        self.tarifa = XmlField('Tarifa')
        self.CodigoTarifaProducto = XmlField('CodigoTarifaProducto')
        self.AltaMedidaEnBaja = AltaMedidaEnBaja()
        super(Tarifa, self).__init__('Tarifa', 'tarifa')

# 2.2


class AltaMedidaEnBaja(XmlModel):

    _sort_order = ('altamedidaenbaja', 'marcamedidaconperdidas', 'kvastrafo', 'porcentajeperdidaspactadas')

    def __init__(self):
        self.altamedidaenbaja = XmlField('AltaMedidaEnBaja')
        self.marcamedidaconperdidas = XmlField('MarcaMedidaConPerdidas')
        self.kvastrafo = XmlField('KVAsTrafo')
        self.porcentajeperdidaspactadas = XmlField('PorcentajePerdidasPactadas')
        super(AltaMedidaEnBaja, self).__init__('AltaMedidaEnBaja', 'altamedidaenbaja')

# 2.2


class Agrupacion(XmlModel):

    _sort_order = ('agrupacion', 'codigoagrupacion')

    def __init__(self):
        self.agrupacion = XmlField('Agrupacion')
        self.codigoagrupacion = XmlField('CodigoAgrupacion')
        super(Agrupacion, self).__init__('Agrupacion', 'agrupacion')

# 2.2


class UtilitiesMedida(XmlModel):

    _sort_order = ('utilitiesmedida', 'medidassobreequipo')

    def __init__(self):
        self.utilitiesmedida = XmlField('UtilitiesMedida')
        self.medidassobreequipo = MedidasSobreEquipo()
        super(UtilitiesMedida, self).__init__('UtilitiesMedida', 'utilitiesmedida')

# 2.2


class MedidasSobreEquipo(XmlModel):

    _sort_order = ('medidassobreequipo', 'medidasobreequipo')

    def __init__(self):
        self.medidassobreequipo = XmlField('MedidasSobreEquipo')
        self.medidasobreequipo = []
        super(MedidasSobreEquipo, self).__init__('MedidasSobreEquipo', 'medidassobreequipo')

# 2.2


class MedidaSobreEquipo(XmlModel):
    _sort_order = ('medidasobreequipo', 'tipomedida', 'lineamedidacontador', 'numeroserie', 'codigodh',
                   'magnitud', 'constantemultiplicadora', 'lecturadesde', 'fechahoradesde', 'lecturahasta',
                   'fechahorahasta', 'tipodelecturadesde', 'tipodelecturahasta', 'consumoleido',
                   'unidadconsumoenergialeido', 'ajuste', 'consumocalculado', 'unidadconsumocalculado')

    def __init__(self):
        self.medidasobreequipo = XmlField('MedidaSobreEquipo')
        self.tipomedida = XmlField('TipoMedida')
        self.lineamedidacontador = XmlField('LineaMedidaContador')
        self.numeroserie = XmlField('NumeroSerie')
        self.codigodh = XmlField('CodigoDH')
        self.magnitud = XmlField('Magnitud')
        self.constantemultiplicadora = XmlField('ConstanteMultiplicadora')
        self.lecturadesde = XmlField('LecturaDesde')
        self.fechahoradesde = XmlField('FechaHoraDesde')
        self.lecturahasta = XmlField('LecturaHasta')
        self.fechahorahasta = XmlField('FechaHoraHasta')
        self.tipodelecturadesde = XmlField('TipoDeLecturaDesde')
        self.tipodelecturahasta = XmlField('TipoDeLecturaHasta')
        self.consumoleido = XmlField('ConsumoLeido')
        self.unidadconsumoenergialeido = XmlField('UnidadConsumoEnergiaLeido')
        self.ajuste = XmlField('Ajuste')
        self.consumocalculado = XmlField('ConsumoCalculado')
        self.unidadconsumocalculado = XmlField('UnidadConsumoCalculado')
        super(MedidaSobreEquipo, self).__init__('MedidaSobreEquipo', 'medidasobreequipo')

# 2.2


class UtilitiesDesgloseConceptosFactura(XmlModel):

    _sort_order = ('utilitiesdesgloseconceptosfactura', 'desgloseconceptosfactura', 'impuestos')

    def __init__(self):
        self.utilitiesdesgloseconceptosfactura = XmlField('UtilitiesDesgloseConceptosFactura')
        self.desgloseconceptosfactura = []
        self.impuestos = Impuestos()
        super(UtilitiesDesgloseConceptosFactura, self).__init__('UtilitiesDesgloseConceptosFactura', 'utilitiesdesgloseconceptosfactura')

# 2.2


class DesgloseConceptosFactura(XmlModel):

    _sort_order = ('desgloseconceptosfactura', 'conceptofacturado', 'tipoconceptofacturado', 'detalleperiodo')

    def __init__(self):
        self.desgloseconceptosfactura = XmlField('DesgloseConceptosFactura')
        self.conceptofacturado = XmlField('ConceptoFacturado')
        self.tipoconceptofacturado = XmlField('TipoConceptoFacturado')
        self.detalleperiodo = []

        super(DesgloseConceptosFactura, self).__init__('DesgloseConceptosFactura', 'desgloseconceptosfactura')

# 2.2

class DetallePeriodo(XmlModel):

    _sort_order = ('detalleperiodo', 'periododh', 'cantidad', 'unidadescantidad', 'fechadesde',
                   'fechahasta', 'periodo', 'unidadperiodo', 'preciounitario', 'unidadespreciounitario',
                   'lineasmedidacontador', 'importe')

    def __init__(self):
        self.detalleperiodo = XmlField('DetallePeriodo')
        self.periododh = XmlField('PeriodoDH')
        self.cantidad = XmlField('Cantidad')
        self.unidadescantidad = XmlField('UnidadesCantidad')
        self.fechadesde = XmlField('FechaDesde')
        self.fechahasta = XmlField('FechaHasta')
        self.periodo = XmlField('Periodo')
        self.unidadperiodo = XmlField('UnidadPeriodo')
        self.preciounitario = XmlField('PrecioUnitario')
        self.unidadespreciounitario = XmlField('UnidadesPrecioUnitario')
        self.lineasmedidacontador = LineasMedidaContador()
        self.importe = XmlField('Importe')
        super(DetallePeriodo, self).__init__('DetallePeriodo', 'detalleperiodo')

# 2.2


class LineasMedidaContador(XmlModel):

    _sort_order = ('lineasmedidacontador', 'lineamedidacontador')

    def __init__(self):
        self.lineasmedidacontador = XmlField('LineasMedidaContador')
        self.lineamedidacontador = []
        super(LineasMedidaContador, self).__init__('LineasMedidaContador', 'lineasmedidacontador', drop_empty=False)

# 2.2


class LineaMedidaContador(XmlModel):

    _sort_order = ('lineamedidacontador',)

    def __init__(self):
        self.lineamedidacontador = XmlField('LineaMedidaContador')
        super(LineaMedidaContador, self).__init__('LineaMedidaContador', 'lineamedidacontador', drop_empty=False)

# 2.2


class Impuestos(XmlModel):

    _sort_order = ('impuestos', 'impuesto')

    def __init__(self):
        self.impuestos = XmlField('Impuestos')
        self.impuesto = []
        super(Impuestos, self).__init__('Impuestos', 'impuestos')

# 2.2


class UtilitiesHistoricoConsumos(XmlModel):

    _sort_order = ('utilitieshistoricoconsumos', 'historicoconsumo')

    def __init__(self):
        self.utilitieshistoricoconsumos = XmlField('UtilitiesHistoricoConsumos')
        self.historicoconsumo = []
        super(UtilitiesHistoricoConsumos, self).__init__('UtilitiesHistoricoConsumos', 'utilitieshistoricoconsumos')

# 2.2


class HistoriCoconsumo(XmlModel):
    _sort_order = ('historicoconsumo', 'periodo', 'descripcion', 'valor', 'unidadmedida',
                   'fechadesdeperiodo', 'fechahastaperiodo')

    def __init__(self):
        self.historicoconsumo = XmlField('HistoricoConsumo')
        self.periodo = XmlField('Periodo')
        self.descripcion = XmlField('Descripcion')
        self.valor = XmlField('Valor')
        self.unidadmedida = XmlField('UnidadMedida')
        self.fechadesdeperiodo = XmlField('FechaDesdePeriodo')
        self.fechahastaperiodo = XmlField('FechaHastaPeriodo')
        super(HistoriCoconsumo, self).__init__('HistoriCoconsumo', 'historicoconsumo')

# 2.2


class UtilitiesMensajes(XmlModel):

    _sort_order = ('utilitiesmensajes', 'listamensajes')

    def __init__(self):
        self.utilitiesmensajes = XmlField('UtilitiesMensajes')
        self.listamensajes = ListaMensajes()
        super(UtilitiesMensajes, self).__init__('UtilitiesMensajes', 'utilitiesmensajes')

# 2.2


class ListaMensajes(XmlModel):

    _sort_order = ('listamensajes', 'mensaje')

    def __init__(self):
        self.listamensajes = XmlField('ListaMensajes')
        self.mensaje = []
        super(ListaMensajes, self).__init__('ListaMensajes', 'listamensajes')

# 2.2


class Mensaje(XmlModel):

    _sort_order = ('mensaje', 'mensajeid', 'contenido')

    def __init__(self):
        self.mensaje = XmlField('Mensaje')
        self.mensajeid = XmlField('MensajeID')
        self.contenido = XmlField('Contenido')
        super(Mensaje, self).__init__('Mensaje', 'mensaje')

# 2.2


class UtilitiesAtrAsociado(XmlModel):

    _sort_order = ('utilitiesatrasociado', 'tarifadeaccesoatr', 'contratodeacceso')

    def __init__(self):
        self.utilitiesatrasociado = XmlField('UtilitiesAtrAsociado')
        self.tarifadeaccesoatr = XmlField('TarifaDeAccesoATR')
        self.contratodeacceso = XmlField('ContratoDeAcceso')
        super(UtilitiesAtrAsociado, self).__init__('UtilitiesAtrAsociado', 'utilitiesatrasociado')

# 2.2


class UtilitiesCurva(XmlModel):

    _sort_order = ('utilitiescurva', 'curvadiaria')

    def __init__(self):
        self.utilitiescurva = XmlField('UtilitiesCurva')
        self.curvadiaria = []
        super(UtilitiesCurva, self).__init__('UtilitiesCurva', 'utilitiescurva')

# 2.2


class CurvaDiaria(XmlModel):

    _sort_order = ('curvadiaria', 'fechacurva', 'datoscurva')

    def __init__(self):
        self.curvadiaria = XmlField('CurvaDiaria')
        self.fechacurva = XmlField('FechaCurva')
        self.datoscurva = DatosCurva()
        super(CurvaDiaria, self).__init__('CurvaDiaria', 'curvadiaria')

# 2.2


class DatosCurva(XmlModel):

    _sort_order = ('datoscurva', 'tipologiacurva', 'tipovalor', 'periodicidad', 'valorcurva')

    def __init__(self):
        self.datoscurva = XmlField('DatosCurva')
        self.tipologiacurva = XmlField('TipologiaCurva')
        self.tipovalor = XmlField('TipoValor')
        self.periodicidad = XmlField('Periodicidad')
        self.valorcurva = XmlField('ValorCurva')
        super(DatosCurva, self).__init__('DatosCurva', 'datoscurva')

# 2.2


















# -*- coding: utf-8 -*-

import sys
sys.path.insert(0, '.')

from facturae import facturae
import signxml
import specs as test_data

import logging
#logging.basicConfig(level=logging.DEBUG)


def clean_cert_str(cert):
    """
    Clean'up an string representation of a certificate
    """
    return cert.replace("-----BEGIN CERTIFICATE-----","").\
                replace("-----END CERTIFICATE-----","").\
                replace("\n","")


with description('Invoice'):
    with before.each:
        self.invoice = facturae.FacturaeRoot()

        header = None
        parties = None
        invoices = None

        self.invoice.feed({
            'fileheader': header,
            'parties': parties,
            'invoices': invoices,
        })

        assert test_data.CERTIFICATE and test_data.CERTIFICATE_PASSWD and test_data.CERTIFICATE_PUBLIC
        self.certificate = open(test_data.CERTIFICATE).read()
        self.certificate_public = clean_cert_str(open(test_data.CERTIFICATE_PUBLIC).read())
        self.password = test_data.CERTIFICATE_PASSWD

    with context('sign'):
        with it('must work as expected'):
            signed = self.invoice.sign(self.certificate, self.password)

            # [!] Public component of provided cert must be inside the signed XML
            assert self.certificate_public in signed, "Expected current certificate '{}' must be inside the signed Invoice".format(test_data.CERTIFICATE_PUBLIC)

            # [!] Sign verification must be accomplished, but must raise
            #  InvalidCertificate for this self-signed cert test
            works = False
            try:
                assert self.invoice.sign_verify(signed)
                works = True
            except signxml.exceptions.InvalidCertificate as invalid_cert:
                # Handle self-signed certs
                if invalid_cert[0].message[2] == "self signed certificate":
                    works = True
            except Exception as e:
                raise e
                works = False

            assert works, "Sign verification must be accomplished"



# -*- coding: utf-8 -*-

from OpenSSL import crypto

class FacturaeUtils(object):

    @staticmethod
    def extract_from_pkcs12(pk, passwd):
        """
        Return the key and the cert from a PKCS12
        """

        assert pk, "PKCS12 must be provided"
        assert passwd, "Passwd must be provided"

        try:
            p12 = crypto.load_pkcs12(pk, passwd)

            priv_key = crypto.dump_privatekey(
                crypto.FILETYPE_PEM, p12.get_privatekey()
            )

            cert = crypto.dump_certificate(
                crypto.FILETYPE_PEM, p12.get_certificate()
            )

            return priv_key, cert

        except Exception as e:
            pass



# -*- coding: utf-8 -*-
from expects import *
from facturae.facturae_parser import FacturaeParser

with description('Facturae Invoice'):
    with before.each:
        with open('./specs/assets/facturae.xsig', 'rb') as f:
            xml_data = f.read()
            self.facturae = FacturaeParser(xml_data)

    with context('parse'):
        with it('worked as expected'):
            expect(self.facturae.sollicitud).to(equal('F19001666A29446424'))
            expect(self.facturae.vat_source).to(equal('A29446424'))
            expect(self.facturae.vat_destination).to(equal('B51065928'))
            expect(self.facturae.num_factures).to(equal('1'))
            expect(self.facturae.total_factures).to(equal('92.83'))
            expect(self.facturae.factures[0]['InvoiceNumber']).to(equal('F19001666'))


    with before.each:
        with open('./specs/assets/facturae_3_2_2.xsig', 'rb') as f:
            xml_data = f.read()
            self.facturae = FacturaeParser(xml_data)

    with context('parse'):
        with it('worked as expected'):
            expect(self.facturae.sollicitud).to(equal('F19001666A29446424'))
            expect(self.facturae.vat_source).to(equal('A29446424'))
            expect(self.facturae.vat_destination).to(equal('B51065928'))
            expect(self.facturae.num_factures).to(equal('1'))
            expect(self.facturae.total_factures).to(equal('92.83'))
            expect(self.facturae.factures[0]['InvoiceNumber']).to(equal('F19001666'))


# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from facturae import __version__

with open('requirements.txt', 'r') as f:
    INSTALL_REQUIRES = f.readlines()

with open('requirements-dev.txt', 'r') as f:
    TESTS_REQUIRES = f.readlines()

setup(
    name='facturae',
    version=__version__,
    url='http://www.gisce.net',
    author='GISCE-TI, S.L.',
    author_email='devel@gisce.net',
    original_author='Electrica Sollerense, S.A.U.',
    original_author_email='informatica@el-gas.es',
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRES,
    license='GPLv3',
    description='Facturae',
    long_description=open('README.md').read(),
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
    ]
)



# -*- coding: utf-8 -*-
from functools import reduce
from lxml import objectify


class FacturaeParser(object):

    def __init__(self, xml_data):
        """Construir Facturae Parser"""
        self.xml_data = xml_data
        try:
            self.xml_obj = objectify.fromstring(self.xml_data)
            self.xml_dict = self.parse_xml()
            self.sollicitud = self.xml_dict.get('BatchIdentifier', False)
            self.num_factures = self.xml_dict.get('InvoicesCount', False)
            self.total_factures = self.xml_dict.get('TotalInvoicesAmount', False)
            self.total_executable = self.xml_dict.get('TotalExecutableAmount', False)
            self.seller = self.xml_dict.get('seller', False)
            self.buyer = self.xml_dict.get('buyer', False)
            self.issuer_type = self.xml_dict.get('InvoiceIssuerType')
            self.vat_source = self._get_from_dict(self.xml_dict, ['seller', 'TaxIdentificationNumber'])
            self.vat_destination = self._get_from_dict(self.xml_dict, ['buyer', 'TaxIdentificationNumber'])
            self.factures = [invoice for invoice in self.xml_dict.get('Invoices')]
        except:
            print('Something went really wrong.')

    def parse_xml(self):
        res = {}

        res.update(self.get_header_dict(self.xml_obj))
        res.update(self.get_parties_dict(self.xml_obj))
        res.update(self.get_invoices_dict(self.xml_obj))

        return res

    def get_header_dict(self, xml_obj):
        res = {}

        header = xml_obj.find('FileHeader')
        batch = header.find('Batch')

        if header is not None:
            res.update({
                'SchemaVersion': header.SchemaVersion.text if header.find(
                    'SchemaVersion') is not None else False,
                'Modality': header.Modality.text if header.find(
                    'Modality') is not None else False,
                'InvoiceIssuerType': header.InvoiceIssuerType.text if header.find(
                    'InvoiceIssuerType') is not None else False,
            })

        if batch is not None:
            res.update({
                'BatchIdentifier': batch.BatchIdentifier.text if batch.find(
                    'BatchIdentifier') is not None else False,
                'InvoicesCount': batch.InvoicesCount.text if batch.find(
                    'InvoicesCount') is not None else False,
                'TotalInvoicesAmount': batch.TotalInvoicesAmount.TotalAmount.text if batch.find(
                    'TotalInvoicesAmount') is not None else False,
                'TotalOutstandingAmount': batch.TotalOutstandingAmount.TotalAmount.text if batch.find(
                    'TotalOutstandingAmount') is not None else False,
                'TotalExecutableAmount': batch.TotalExecutableAmount.TotalAmount.text if batch.find(
                    'TotalExecutableAmount') is not None else False,
                'InvoiceCurrencyCode': batch.InvoiceCurrencyCode.text if batch.find(
                    'InvoiceCurrencyCode') is not None else False,
            })

        return res

    def get_parties_dict(self, xml_obj):
        res = {'seller': {}, 'buyer': {}}
        parties = xml_obj.find('Parties')

        if parties is not None:
            seller = parties.SellerParty
            buyer = parties.BuyerParty

            if seller is not None:
                res['seller'] = self._get_party_data(seller)

            if buyer is not None:
                res['buyer'] = self._get_party_data(buyer)

        return res

    def _get_party_data(self, party):
        res = {}

        tax_identification = party.find('TaxIdentification')
        legal_entity = party.find('LegalEntity')
        individual = party.find('Individual')
        party_address = None

        if tax_identification is not None:
            res.update({
                'PersonTypeCode': tax_identification.PersonTypeCode.text if tax_identification.find(
                    'PersonTypeCode') is not None else False,
                'ResidenceTypeCode': tax_identification.ResidenceTypeCode.text if tax_identification.find(
                    'ResidenceTypeCode') is not None else False,
                'TaxIdentificationNumber': tax_identification.TaxIdentificationNumber.text if tax_identification.find(
                    'TaxIdentificationNumber') is not None else False,
            })

        if legal_entity is not None:
            res.update({
                'CorporateName': legal_entity.CorporateName.text if legal_entity.find(
                    'CorporateName') is not None else False,
                'TradeName': legal_entity.TradeName.text if legal_entity.find(
                    'TradeName') is not None else False,
            })

            party_address = legal_entity.find('AddressInSpain')

        if individual is not None:
            res.update({
                'Name': individual.Name.text if individual.find(
                    'Name') is not None else False,
                'FirstSurname': individual.FirstSurname.text if individual.find(
                    'FirstSurname') is not None else False,
                'SecondSurname': individual.SecondSurname.text if individual.find(
                    'SecondSurname') is not None else False,
            })

            party_address = individual.find('AddressInSpain')

        if party_address is not None:
            res.update({
                'Address': {
                    'Street': party_address.Address.text if party_address.find(
                        'Address') is not None else False,
                    'PostCode': party_address.PostCode.text if party_address.find(
                        'PostCode') is not None else False,
                    'Town': party_address.Town.text if party_address.find(
                        'Town') is not None else False,
                    'Province': party_address.Province.text if party_address.find(
                        'Province') is not None else False,
                    'CountryCode': party_address.CountryCode.text if party_address.find(
                        'CountryCode') is not None else False,
                }
            })

        return res

    def get_invoices_dict(self, xml_obj):
        res = {'Invoices': []}
        invoices = xml_obj.find('Invoices')

        for invoice in invoices.findall('Invoice'):
            invoice_res = {}

            invoice_header = invoice.find('InvoiceHeader')
            if invoice_header is not None:
                invoice_res.update({
                    'InvoiceNumber': invoice_header.InvoiceNumber.text if invoice_header.find(
                        'InvoiceNumber') is not None else False,
                    'InvoiceSeriesCode': invoice_header.InvoiceSeriesCode.text if invoice_header.find(
                        'InvoiceSeriesCode') is not None else False,
                    'InvoiceDocumentType': invoice_header.InvoiceDocumentType.text if invoice_header.find(
                        'InvoiceDocumentType') is not None else False,
                    'InvoiceClass': invoice_header.InvoiceClass.text if invoice_header.find(
                        'InvoiceClass') is not None else False,
                })

            invoice_issue_data = invoice.find('InvoiceIssueData')
            if invoice_issue_data is not None:
                invoice_res.update({
                    'IssueDate': invoice_issue_data.IssueDate.text if invoice_issue_data.find(
                        'IssueDate') is not None else False,
                    'InvoiceCurrencyCode': invoice_issue_data.InvoiceCurrencyCode.text if invoice_issue_data.find(
                        'InvoiceCurrencyCode') is not None else False,
                    'TaxCurrencyCode': invoice_issue_data.TaxCurrencyCode.text if invoice_issue_data.find(
                        'TaxCurrencyCode') is not None else False,
                    'LanguageName': invoice_issue_data.LanguageName.text if invoice_issue_data.find(
                        'LanguageName') is not None else False,
                })

            invoice_res.update({
                'Taxes': self._get_taxes(invoice.TaxesOutputs) if invoice.find('TaxesOutputs') is not None else False,
            })

            invoice_totals = invoice.find('InvoiceTotals')
            if invoice_totals is not None:
                invoice_res.update({
                    'TotalGrossAmount': invoice_totals.TotalGrossAmount.text if invoice_totals.find(
                        'TotalGrossAmount') is not None else False,
                    'TotalGrossBeforeTaxes': invoice_totals.TotalGrossAmountBeforeTaxes.text if invoice_totals.find(
                        'TotalGrossAmountBeforeTaxes') is not None else False,
                    'TotalTaxOutputs': invoice_totals.TotalTaxOutputs.text if invoice_totals.find(
                        'TotalTaxOutputs') is not None else False,
                    'TotalTaxesWithheld': invoice_totals.TotalTaxesWithheld.text if invoice_totals.find(
                        'TotalTaxesWithheld') is not None else None,
                    'InvoiceTotal': invoice_totals.InvoiceTotal.text if invoice_totals.find(
                        'InvoiceTotal') is not None else False,
                    'TotalOutstandingAmount': invoice_totals.TotalOutstandingAmount.text if invoice_totals.find(
                        'TotalOutstandingAmount') is not None else False,
                    'AmountsWithheld': self._get_amounts_with_held(invoice_totals.AmountsWithheld) if invoice_totals.find(
                        'AmountsWithheld') is not None else None,
                    'TotalExecutableAmount': invoice_totals.TotalExecutableAmount.text if invoice_totals.find(
                        'TotalExecutableAmount') is not None else False,
                })

            invoice_res.update({
                'InvoiceLines': self._get_invoice_lines(invoice.Items) if invoice.find(
                    'Items') is not None else False,
            })

            payment_details = invoice.find('PaymentDetails')
            if payment_details is not None:
                installment = payment_details.find('Installment')
                if installment is not None:
                    invoice_res.update({
                        'InstallmentDueDate': installment.InstallmentDueDate.text if installment.find(
                            'InstallmentDueDate') is not None else False,
                        'InstallmentAmount': installment.InstallmentAmount.text if installment.find(
                            'InstallmentAmount') is not None else False,
                        'PaymentMeans': installment.PaymentMeans.text if installment.find(
                            'PaymentMeans') is not None else False,
                    })

                    bank_account = installment.find('AccountToBeDebited')
                    if bank_account is not None:
                        invoice_res.update({
                            'IBAN': bank_account.IBAN.text if bank_account.find(
                                'IBAN') is not None else False,
                            'BIC': bank_account.BIC.text if bank_account.find(
                                'BIC') is not None else False,
                        })

            additional_data = invoice.find('AdditionalData')
            if additional_data is not None:
                invoice_res.update({
                    'Attachments': self._get_attachments(additional_data.RelatedDocuments) if additional_data.find(
                        'RelatedDocuments') is not None else False,
                    'AdditionalInformation': additional_data.InvoiceAdditionalInformation.text if additional_data.find(
                        'InvoiceAdditionalInformation') is not None else False,
                })

            res['Invoices'].append(invoice_res)

        return res

    def _get_amounts_with_held(self, amountwhitheld):

        amounts_with_held = {
            'WithholdingReason': amountwhitheld.WithholdingReason.text if amountwhitheld.find(
                'WithholdingReason') is not None else None,
            'WithholdingRate': amountwhitheld.WithholdingRate.text if amountwhitheld.find(
                'WithholdingRate') is not None else None,
            'WithholdingAmount': amountwhitheld.WithholdingAmount.text if amountwhitheld.find(
                'WithholdingAmount') is not None else None
        }

        return amounts_with_held

    def _get_taxes(self, taxes):
        res = []

        for tax in taxes.findall('Tax'):
            tax_res = {
                'TaxTypeCode': tax.TaxTypeCode.text if tax.find(
                    'TaxTypeCode') is not None else False,
                'TaxRate': tax.TaxRate.text if tax.find(
                    'TaxRate') is not None else False,
                'TaxableBase': tax.TaxableBase.TotalAmount.text if tax.find(
                    'TaxableBase') is not None else False,
                'TaxAmount': tax.TaxAmount.TotalAmount.text if tax.find(
                    'TaxAmount') is not None else False,
            }
            res.append(tax_res)

        return res

    def _get_invoice_lines(self, items):
        res = []

        for line in items.findall('InvoiceLine'):
            line_res = {
                'ItemDescription': line.ItemDescription.text if line.find(
                    'ItemDescription') is not None else False,
                'Quantity': line.Quantity.text if line.find(
                    'Quantity') is not None else False,
                'UnitPriceWithoutTax': line.UnitPriceWithoutTax.text if line.find(
                    'UnitPriceWithoutTax') is not None else False,
                'TotalCost': line.TotalCost.text if line.find(
                    'TotalCost') is not None else False,
                'GrossAmount': line.GrossAmount.text if line.find(
                    'GrossAmount') is not None else False,
                'TaxesOutputs': self._get_taxes(line.TaxesOutputs) if line.find(
                    'TaxesOutputs') is not None else False,
            }
            res.append(line_res)

        return res

    def _get_attachments(self, related_documents):
        res = []

        for attachment in related_documents.findall('Attachment'):
            attachment_res = {
                'AttachmentCompressionAlgorithm': attachment.AttachmentCompressionAlgorithm.text if attachment.find(
                    'AttachmentCompressionAlgorithm') is not None else False,
                'AttachmentFormat': attachment.AttachmentFormat.text if attachment.find(
                    'AttachmentFormat') is not None else False,
                'AttachmentEncoding': attachment.AttachmentEncoding.text if attachment.find(
                    'AttachmentEncoding') is not None else False,
                'AttachmentData': attachment.AttachmentData.text if attachment.find(
                    'AttachmentData') is not None else False,
            }
            res.append(attachment_res)

        return res

    def _get_from_dict(self,dataDict, mapList):
        """Iterate nested dictionary"""
        return reduce(dict.get, mapList, dataDict)

